/*
 * Copyright (c) 2016, 2017, 2018, 2019 FabricMC
 * Copyright 2023 The Quilt Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package net.fabricmc.fabric.impl.recipe.ingredient.builtin;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import com.mojang.brigadier.exceptions.CommandSyntaxException;
import com.mojang.datafixers.util.Either;
import com.mojang.serialization.Codec;
import com.mojang.serialization.DataResult;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import org.jetbrains.annotations.Nullable;
import net.fabricmc.fabric.api.recipe.v1.ingredient.CustomIngredient;
import net.fabricmc.fabric.api.recipe.v1.ingredient.CustomIngredientSerializer;
import net.minecraft.class_1799;
import net.minecraft.class_1856;
import net.minecraft.class_2487;
import net.minecraft.class_2512;
import net.minecraft.class_2522;
import net.minecraft.class_2540;
import net.minecraft.class_2960;
import net.minecraft.class_5699;

public class NbtIngredient implements CustomIngredient {
	public static final CustomIngredientSerializer<NbtIngredient> SERIALIZER = new Serializer();

	private final class_1856 base;
	@Nullable
	private final class_2487 nbt;
	private final boolean strict;

	public NbtIngredient(class_1856 base, @Nullable class_2487 nbt, boolean strict) {
		if (nbt == null && !strict) {
			throw new IllegalArgumentException("NbtIngredient can only have null NBT in strict mode");
		}

		this.base = base;
		this.nbt = nbt;
		this.strict = strict;
	}

	@Override
	public boolean test(class_1799 stack) {
		if (!base.method_8093(stack)) return false;

		if (strict) {
			return Objects.equals(nbt, stack.method_7969());
		} else {
			return class_2512.method_10687(nbt, stack.method_7969(), true);
		}
	}

	@Override
	public List<class_1799> getMatchingStacks() {
		List<class_1799> stacks = new ArrayList<>(List.of(base.method_8105()));
		stacks.replaceAll(stack -> {
			class_1799 copy = stack.method_7972();

			if (nbt != null) {
				copy.method_7980(nbt.method_10553());
			}

			return copy;
		});
		stacks.removeIf(stack -> !base.method_8093(stack));
		return stacks;
	}

	@Override
	public boolean requiresTesting() {
		return true;
	}

	@Override
	public CustomIngredientSerializer<?> getSerializer() {
		return SERIALIZER;
	}

	private class_1856 getBase() {
		return base;
	}

	private class_2487 getNbt() {
		return nbt;
	}

	private boolean isStrict() {
		return strict;
	}

	private static class Serializer implements CustomIngredientSerializer<NbtIngredient> {
		private static final class_2960 ID = new class_2960("fabric", "nbt");

		// Supports decoding the NBT as a string as well as the object.
		private static final Codec<class_2487> NBT_CODEC = class_5699.method_33817(
				Codec.STRING, class_2487.field_25128
		).flatXmap(either -> either.map(s -> {
			try {
				return DataResult.success(class_2522.method_10718(s));
			} catch (CommandSyntaxException e) {
				return DataResult.error(e::getMessage);
			}
		}, DataResult::success), nbtCompound -> DataResult.success(Either.left(nbtCompound.method_10714())));

		private static final Codec<NbtIngredient> ALLOW_EMPTY_CODEC = createCodec(class_1856.field_46095);
		private static final Codec<NbtIngredient> DISALLOW_EMPTY_CODEC = createCodec(class_1856.field_46096);

		private static Codec<NbtIngredient> createCodec(Codec<class_1856> ingredientCodec) {
			return RecordCodecBuilder.create(instance ->
					instance.group(
							ingredientCodec.fieldOf("base").forGetter(NbtIngredient::getBase),
							NBT_CODEC.optionalFieldOf("nbt", null).forGetter(NbtIngredient::getNbt),
							Codec.BOOL.optionalFieldOf("strict", false).forGetter(NbtIngredient::isStrict)
					).apply(instance, NbtIngredient::new)
			);
		}

		@Override
		public class_2960 getIdentifier() {
			return ID;
		}

		@Override
		public Codec<NbtIngredient> getCodec(boolean allowEmpty) {
			return allowEmpty ? ALLOW_EMPTY_CODEC : DISALLOW_EMPTY_CODEC;
		}

		@Override
		public NbtIngredient read(class_2540 buf) {
			class_1856 base = class_1856.method_8086(buf);
			class_2487 nbt = buf.method_10798();
			boolean strict = buf.readBoolean();
			return new NbtIngredient(base, nbt, strict);
		}

		@Override
		public void write(class_2540 buf, NbtIngredient ingredient) {
			ingredient.base.method_8088(buf);
			buf.method_10794(ingredient.nbt);
			buf.method_52964(ingredient.strict);
		}
	}
}
